/**
 * Copyright (C) 2011 - 101loops.com <dev@101loops.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.crashnote.core.config;

import com.crashnote.core.log.LogLog;
import com.crashnote.core.util.ClassUtil;

import java.util.*;

import static com.crashnote.core.config.Config.*;
import static com.crashnote.core.config.ConfigBase.LIB_URL_BOARD;
import static com.crashnote.core.config.ConfigBase.getConfigKey;

/**
 * This class is responsible for instantiating a {@link Config} object and internally apply
 * the configuration settings (e.g. from system props, property file).
 * <p/>
 * Finally it validates the config ("fail fast") and returns it.
 */
public class ConfigFactory<C extends Config> {

    protected final C config;
    private final LogLog logger;

    /**
     * specifies whether the config object was already validated
     */
    private boolean validated = false;

    /**
     * indicates whether the external configuration sources are already processed
     */
    private boolean extConfigLoaded = false;

    /**
     * name of the configuration file to parse for settings
     */
    public static final String PROP_FILE = ConfigBase.LIB_NAME.toLowerCase() + ".properties";

    // SETUP ======================================================================================

    public ConfigFactory(final C config) {
        this.config = config;
        this.logger = config.getLogger(this.getClass());
    }

    // INTERFACE ==================================================================================

    public C get() {
        if (!validated) {
            prepare();
            validate(); // validate first to fail fast if necessary
            validated = true;
        }
        return (C) config.copy();
    }

    // SHARED =====================================================================================

    protected void prepare() {

        // load external configuration (e.g. property file)
        loadExt();
    }

    protected void validate() {
        if (isEnabled())
            validateKey();
        else
            logger.info("transfer to Cloud is DISABLED");
    }

    protected void loadSystemProperties() {
        applyProperties(System.getProperties(), true);
    }

    protected void loadEnvProperties() {
        final Map<String, String> env = System.getenv();
        final Properties props = new Properties();
        for (final String name : env.keySet()) {
            props.setProperty(name, env.get(name));
        }
        applyProperties(props, true);
    }

    protected void loadFileProperties() {
        applyProperties(new ClassUtil().load(PROP_FILE), false);
    }

    protected void applyProperties(final Properties props, final boolean strict) {

        final String key = getProperty(props, PROP_API_KEY, strict);
        if (key != null) setKey(key);

        final String debug = getProperty(props, PROP_DEBUG, strict);
        if (debug != null) setDebug(debug);

        final String mode = getProperty(props, PROP_SYNC, strict);
        if (mode != null) setSync(mode);

        final String enabled = getProperty(props, PROP_ENABLED, strict);
        if (enabled != null) setEnabled(enabled);

        final String host = getProperty(props, PROP_API_HOST, strict);
        if (host != null) setHost(host);

        final String secure = getProperty(props, PROP_API_SECURE, strict);
        if (secure != null) setSecure(secure);

        final String port = getProperty(props, PROP_API_PORT, strict);
        if (port != null) setPort(port);

        final String sslPort = getProperty(props, PROP_API_PORT_SSL, strict);
        if (sslPort != null) setSslPort(sslPort);
    }

    protected final String getProperty(final Properties props, final String name, final boolean strict) {
        return props.getProperty(getConfigKey(name),
            props.getProperty(getConfigKey(name, '_'),
                props.getProperty(getConfigKey(name, '-'),
                    strict ? null : props.getProperty(name, null))));
    }

    protected final void loadExt() {
        if (!extConfigLoaded) {
            loadFileProperties(); // #1: load property file
            loadEnvProperties(); // #2: load environment props
            loadSystemProperties(); // #3: load system properties
            extConfigLoaded = true;
        }
    }

    // INTERNALS ==================================================================================

    private boolean isEnabled() {
        if (config.isEnabled()) {
            logger.info("Status: ON");
            return true;
        } else {
            logger.info("Status: OFF");
            return false;
        }
    }

    private void validateKey() {
        final String key = config.getKey();
        if (key == null || key.length() == 0) {
            throw new IllegalArgumentException(
                "The API Key is missing, please login to the web app under [" + LIB_URL_BOARD + "], " +
                    "browse to your app and consult the 'Install' instructions.");

        } else if (key.length() != 36 || key.charAt(8) != '-' || key.charAt(13) != '-' || key.charAt(18) != '-' || key.charAt(23) != '-')
            throw new IllegalArgumentException(
                "The API Key appears to be invalid (it should be 36 characters long with 4 dashes), " +
                    "please login to the web app under [" + LIB_URL_BOARD + "], " +
                    "browse to your app and consult the 'Install' instructions.");
    }

    // SET ========================================================================================

    public void setEnabled(final String enabled) {
        config.setEnabled(enabled);
    }

    public void setDebug(final String debug) {
        config.setDebug(debug);
    }

    public void setSslPort(final String sslPort) {
        config.setSSLPort(sslPort);
    }

    public void setSecure(final String secure) {
        config.setSecure(secure);
    }

    public void setSync(final String sync) {
        config.setSync(sync);
    }

    public void setPort(final String port) {
        config.setPort(port);
    }

    public void setProfile(final String profile) {
        config.setAppProfile(profile);
    }

    public void setAppVersion(final String appVersion) {
        config.setAppVersion(appVersion);
    }

    public void setSecure(final Boolean secure) {
        config.setSecure(secure);
    }

    public void setConnectionTimeout(final String timeout) {
        config.setConnectionTimeout(timeout);
    }

    public void setKey(final String key) {
        config.setKey(key);
    }

    public void setEnabled(final boolean enabled) {
        config.setEnabled(enabled);
    }

    public void setPort(final int port) {
        config.setPort(port);
    }

    public void setSslPort(final int sslPort) {
        config.setSSLPort(sslPort);
    }

    public void setHost(final String host) {
        config.setHost(host);
    }

    public void setDebug(final Boolean debug) {
        config.setDebug(debug);
    }

    // GET ========================================================================================

    public LogLog getLogger() {
        return logger;
    }
}
